@@ -21,7 +21,7 @@ import flask
 from flask import Response, request
 # just for forwarding
 from flask import session, jsonify, redirect, url_for, Blueprint
-from flask import make_response, abort, render_template, current_app
+from flask import make_response, abort, render_template, current_app, g
 
 import logging
 log = logging.getLogger("fsa")
@@ -40,7 +40,8 @@ class AuthException(BaseException):
         self.status = status
 
 
-# special type casts
+#
+# TYPE CASTS
 #
 def bool_cast(s: str) -> Optional[bool]:
     """Parses a bool."""
@@ -55,7 +56,12 @@ def int_cast(s: str) -> Optional[int]:
 
 
 class path(str):
-    """Type to distinguish path parameters."""
+    """Type to distinguish str path parameters."""
+    pass
+
+
+class string(str):
+    """Type to distinguish str path parameters."""
     pass
 
 
@@ -65,12 +71,15 @@ CASTS: Dict[type, Callable[[str], object]] = {
     int: int_cast,
     inspect._empty: str,
     path: str,
+    string: str,
     dt.date: dt.date.fromisoformat,
     dt.time: dt.time.fromisoformat,
     dt.datetime: dt.datetime.fromisoformat
 }
 
-# special group names
+#
+# PREDEFINED GROUP NAMES
+#
 ANY = "ANYone can come in, no authentication required"
 ALL = "ALL authentified users are allowed"
 NONE = "NONE can come in, the path is forbidden"
@@ -78,13 +87,13 @@ NONE = "NONE can come in, the path is forbidden"
 
 def typeof(p: inspect.Parameter):
     """Guess parameter type, possibly with some type inference."""
-    if p.kind == p.VAR_KEYWORD:
+    if p.kind is p.VAR_KEYWORD:
         return dict
-    elif p.kind == p.VAR_POSITIONAL:
+    elif p.kind is p.VAR_POSITIONAL:
         return list
-    elif p.annotation != inspect._empty:
+    elif p.annotation is not inspect._empty:
         return p.annotation
-    elif p.default is not None and p.default != inspect._empty:
+    elif p.default is not None and p.default is not inspect._empty:
         return type(p.default)  # type inference!
     else:
         return str
@@ -93,7 +102,16 @@ def typeof(p: inspect.Parameter):
 class Reference:  # type: Any
     """Convenient object wrapper class.
 
-    The wrapper forwards most method calls to the wrapped object.
+    The wrapper forwards most method calls to the wrapped object, so that
+    the reference can be imported even if the object is not created yet.
+
+    ```python
+    r = Reference()
+    o = …
+    r.set(o)
+    r.whatever(…) # behaves as o.whatever(…)
+    ```
+
     """
 
     def __init__(self, obj: Any = None, set_name: str = "set"):
@@ -103,20 +121,16 @@ class Reference:  # type: Any
         - set_name: provide another name for the "set" function.
         """
         self._obj = None
-        # possibly rename the "set" method
+        # create "set" method, which may use another name…
         if set_name is None:
             set_name = "set"
-        elif set_name != "set":
-            setattr(self, set_name, getattr(self, "set"))
-            # cannot really delete "set"…
-            # delattr(self, "set"): attribute error
-            # delattr(self.__class__, "set"): removes for all instances!
+        setattr(self, set_name, getattr(self, "_set_obj"))
         # keep track of initial methods for later cleanup
         self._init = set(self.__dir__() + ["_init"])
         if obj is not None:
-            getattr(self, set_name)(obj)
+            self._set_obj(obj)
 
-    def set(self, obj):
+    def _set_obj(self, obj):
         """Set current wrapped object, possibly replacing the previous one."""
         log.debug(f"setting reference to {obj} ({type(obj)})")
         self._obj = obj
@@ -131,6 +145,8 @@ class Reference:  # type: Any
 
     # forward standard methods
     # automating that with setattr/getattr does not work…
+    # contrary to the documentation say, it seems that str(obj)
+    # really calls obj.__class__.__str__() and *not* obj.__str__().
     def __str__(self):
         return self._obj.__str__()
 
@@ -193,9 +209,9 @@ class Flask(flask.Flask):
 
     - the `route` decorator manages authentication, authorization and
       parameters transparently.
-    - several additional methods are provided: `init_app`, `get_user_pass`,
+    - several additional methods are provided: `get_user_pass`,
       `user_in_group`, `check_password`, `hash_password`, `create_token`,
-      `get_user`, `current_user`.
+      `get_user`, `current_user`, `clear_caches`.
     """
 
     def __init__(self, *args, **kwargs):
@@ -205,11 +221,6 @@ class Flask(flask.Flask):
         # overwritten late because called by upper Flask initialization
         self.add_url_rule = self._fsa.add_url_rule
 
-    # forward some methods
-    def init_app(self):
-        """Initialize underlying extension."""
-        return self._fsa.init_app(self)
-
     def get_user_pass(self, gup):
         """Set `get_user_pass` helper function."""
         return self._fsa.get_user_pass(gup)
@@ -219,15 +230,15 @@ class Flask(flask.Flask):
         return self._fsa.user_in_group(uig)
 
     def check_password(self, pwd, ref):
-        """Check whether password is ok wrt to current configuration."""
+        """Check whether password is ok wrt to current password manager."""
         return self._fsa.check_password(pwd, ref)
 
     def hash_password(self, pwd):
-        """Hash password using current password scheme."""
+        """Hash password using current password manager scheme."""
         return self._fsa.hash_password(pwd)
 
-    def create_token(self, user):
-        """Create a token with the current token configuration."""
+    def create_token(self, user: str = None):
+        """Create a token with the current token scheme."""
         return self._fsa.create_token(user)
 
     def get_user(self):
@@ -238,6 +249,16 @@ class Flask(flask.Flask):
         """Get current authenticated user, if any."""
         return self._fsa.current_user()
 
+    def clear_caches(self):
+        """Clear internal caches."""
+        self._fsa.clear_caches()
+
+
+class Resp(Response):
+    """Genenerate a text/plain Response."""
+    def __init__(self, msg: str, code: int):
+        super().__init__(msg, code, content_type="text/plain")
+
 
 # actual class
 class FlaskSimpleAuth:
@@ -250,9 +271,13 @@ class FlaskSimpleAuth:
         self._app = app
         self._get_user_pass = None
         self._user_in_group = None
-        # actual initialization is delayed
+        self._http_auth = None
+        # actual initialization is deferred
         self._initialized = False
 
+    #
+    # HOOKS
+    #
     def _auth_set_user(self):
         """Before request hook to perform early authentication."""
         self._user = None
@@ -265,30 +290,65 @@ class FlaskSimpleAuth:
         try:
             self._user = self.get_user()
         except AuthException as e:
-            return e.message, e.status
+            return Resp(e.message, e.status)
         assert self._user is not None
 
     def _auth_after_cleanup(self, res: Response):
         """After request hook to cleanup authentication and detect missing
         authorization."""
         self._user = None
+        # should it always return 500?
         if res.status_code < 400 and self._need_authorization:
             method, path = request.method, request.path
             log.warning(f"missing authorization on {method} {path}")
             if self._check:
-                return Response("missing authorization check", 500)
+                return Resp("missing authorization check", 500)
+        return res
+
+    # set a cookie if needed and none was sent
+    # we assume that thanks to max_age the client will not send stale cookies
+    def _set_auth_cookie(self, res: Response):
+        if self._carrier == "cookie":
+            assert self._token is not None and self._name is not None
+            if self._user is not None:
+                if self._name in request.cookies:
+                    user, exp = self._get_token_auth_exp(request.cookies[self._name])
+                    # reset token when 25% time remains
+                    limit = self._timestamp(dt.datetime.utcnow() + 0.25 * dt.timedelta(minutes=self._delay))
+                    set_cookie = exp < limit
+                else:
+                    set_cookie = True
+                if set_cookie:
+                    res.set_cookie(self._name, self.create_token(self._user),
+                                   max_age=int(60 * self._delay))
+        return res
+
+    def _set_www_authenticate(self, res: Response):
+        if res.status_code == 401:
+            if self._auth in ("basic", "password"):
+                res.headers["WWW-Authenticate"] = f"Basic realm=\"{self._realm}\""
+            elif self._auth in ("http-basic", "http-digest", "http-token", "digest"):
+                res.headers["WWW-Authenticate"] = self._http_auth.authenticate_header()
+            elif self._auth == "param":
+                # not sure this one makes much sense
+                res.headers["WWW-Authenticate"] = f"Param realm=\"{self._realm}\""
+            elif self._carrier == "bearer":
+                res.headers["WWW-Authenticate"] = f"{self._name} realm=\"{self._realm}\""
         return res
 
     def get_user_pass(self, gup):
         """Set `get_user_pass` helper, can be used as a decorator."""
-        self._get_user_pass = gup
+        self._get_user_pass = functools.lru_cache(maxsize=1024)(gup)
         return gup
 
     def user_in_group(self, uig):
         """Set `user_in_group` helper, can be used as a decorator."""
-        self._user_in_group = uig
+        self._user_in_group = functools.lru_cache(maxsize=1024)(uig)
         return uig
 
+    #
+    # DEFERRED INITIALIZATION
+    #
     def initialize(self):
         """Run late initialization on current app."""
         assert self._app is not None
@@ -305,30 +365,48 @@ class FlaskSimpleAuth:
         self._app = app
         conf = app.config
         #
-        # auth setup
+        # overall auth setup
         #
-        self._auth = conf.get("FSA_TYPE", "httpd")
-        assert self._auth in ("httpd", "none", "fake", "basic", "param", "password", "token")
+        self._auth = conf.get("FSA_AUTH", "httpd")
+        assert self._auth in ("httpd", "none", "fake", "basic", "param", "password",
+                              "token", "http-basic", "http-digest", "http-token", "digest")
         self._lazy = conf.get("FSA_LAZY", True)
         self._always = conf.get("FSA_ALWAYS", True)
         self._check = conf.get("FSA_CHECK", True)
-        # register auth request hooks
-        app.before_request(self._auth_set_user)
-        app.after_request(self._auth_after_cleanup)
         import re
         self._skip_path = [re.compile(r).match for r in conf.get("FSA_SKIP_PATH", [])]
         #
         # token setup
         #
-        self._type = conf.get("FSA_TOKEN_TYPE", "fsa")
-        self._name = conf.get("FSA_TOKEN_NAME", None)
-        realm = conf.get("FSA_TOKEN_REALM", self._app.name).lower()
-        # tr -cd "[a-z0-9_]" "": is there a better way to do that?
-        keep_char = re.compile(r"[-a-z0-9_]").match
-        self._realm = "".join(c for c in realm if keep_char(c))
+        self._token = conf.get("FSA_TOKEN_TYPE", "fsa")
+        if self._token not in (None, "fsa", "jwt"):
+            raise Exception(f"Unexpected token type (FSA_TOKEN_TYPE): {self._token}")
+        # where to put/look for the token
+        need_carrier = self._token is not None
+        self._carrier = conf.get("FSA_TOKEN_CARRIER", "bearer" if need_carrier else None)
+        if self._carrier not in (None, "bearer", "param", "cookie"):
+            raise Exception(f"Unexpected token carrier (FSA_TOKEN_CARRIER): {self._carrier}")
+        # sanity checks
+        if need_carrier and self._carrier is None:
+            raise Exception(f"Token type {self._token} requires a carrier")
+        # name of token for cookie or param
+        default_name = "auth" if self._carrier in ("param", "cookie") else "Bearer"
+        self._name = conf.get("FSA_TOKEN_NAME", default_name)
+        if need_carrier and self._name is None:
+            raise Exception(f"Token carrier {self._carrier} requires a name")
+        # token realm…
+        realm = conf.get("FSA_TOKEN_REALM", self._app.name)
+        if self._token == "fsa":
+            # simplify realm for fsa
+            keep_char = re.compile(r"[-A-Za-z0-9]").match
+            realm = "".join(c if keep_char(c) else "-" for c in realm)
+            realm = "-".join(filter(lambda s: s != "", realm.split("-")))
+        self._realm = realm
+        # time validity
         self._delay = conf.get("FSA_TOKEN_DELAY", 60.0)
         self._grace = conf.get("FSA_TOKEN_GRACE", 0.0)
-        if self._type is not None and self._type == "jwt":
+        # token signature
+        if self._token is not None and self._token == "jwt":
             algo = conf.get("FSA_TOKEN_ALGO", "HS256")
             if algo[0] in ("R", "E", "P"):
                 assert "FSA_TOKEN_SECRET" in conf and "FSA_TOKEN_SIGN" in conf, \
@@ -344,13 +422,13 @@ class FlaskSimpleAuth:
             # list of 94 chars, about 6.5 bits per char
             chars = string.ascii_letters + string.digits + string.punctuation
             self._secret = ''.join(random.SystemRandom().choices(chars, k=40))
-        if self._type is None:
+        if self._token is None:
             pass
-        elif self._type == "fsa":
+        elif self._token == "fsa":
             self._sign = self._secret
             self._algo = conf.get("FSA_TOKEN_ALGO", "blake2s")
             self._siglen = conf.get("FSA_TOKEN_LENGTH", 16)
-        elif self._type == "jwt":
+        elif self._token == "jwt":
             algo = conf.get("FSA_TOKEN_ALGO", "HS256")
             self._algo = algo
             if algo[0] in ("R", "E", "P"):
@@ -363,9 +441,9 @@ class FlaskSimpleAuth:
                 raise Exception(f"unexpected jwt FSA_TOKEN_ALGO ({algo})")
             self._siglen = 0
         else:
-            raise Exception(f"invalid FSA_TOKEN_TYPE ({self._type})")
+            raise Exception(f"invalid FSA_TOKEN_TYPE ({self._token})")
         #
-        # parameters
+        # HTTP parameter names
         #
         self._login = conf.get("FSA_FAKE_LOGIN", "LOGIN")
         self._userp = conf.get("FSA_PARAM_USER", "USER")
@@ -377,6 +455,8 @@ class FlaskSimpleAuth:
         # round. Ident '2y' is same as '2b' but apache compatible.
         scheme = conf.get("FSA_PASSWORD_SCHEME", "bcrypt")
         if scheme is not None:
+            if scheme == "plaintext":
+                log.warning("plaintext password manager is a bad idea")
             options = conf.get("FSA_PASSWORD_OPTIONS",
                                {'bcrypt__default_rounds': 4,
                                 'bcrypt__default_ident': '2y'})
@@ -388,9 +468,35 @@ class FlaskSimpleAuth:
         # hooks
         #
         if "FSA_GET_USER_PASS" in conf:
-            self._get_user_pass = conf["FSA_GET_USER_PASS"]
+            self.get_user_pass(conf["FSA_GET_USER_PASS"])
         if "FSA_USER_IN_GROUP" in conf:
-            self._user_in_group = conf["FSA_USER_IN_GROUP"]
+            self.user_in_group(conf["FSA_USER_IN_GROUP"])
+        #
+        # http auth setup
+        #
+        if self._auth in ("http-basic", "http-digest", "http-token", "digest"):
+            opts = conf.get("FSA_HTTP_AUTH_OPTS", {})
+            import flask_httpauth as fha
+            if self._auth == "http-basic":
+                self._http_auth = fha.HTTPBasicAuth(realm=self._realm, **opts)
+                self._http_auth.verify_password(self._check_password)
+            elif self._auth in ("http-digest", "digest"):
+                self._http_auth = fha.HTTPDigestAuth(realm=self._realm, **opts)
+                # FIXME nonce & opaque callbacks? session??
+            elif self._auth == "http-token":
+                self._http_auth = fha.HTTPTokenAuth(scheme=self._name, realm=self._realm, **opts)
+                self._http_auth.verify_token(self._get_token_auth)
+            self._http_auth.get_password(self._get_user_pass)
+            # FIXME error_handler?
+        else:
+            self._http_auth = None
+        #
+        # register auth request hooks
+        #
+        app.before_request(self._auth_set_user)
+        app.after_request(self._auth_after_cleanup)
+        app.after_request(self._set_auth_cookie)
+        app.after_request(self._set_www_authenticate)
         #
         # blueprint hacks
         #
@@ -443,11 +549,28 @@ class FlaskSimpleAuth:
             log.warning("password authentication over an insecure request")
         ref = self._get_user_pass(user)
         if ref is None:
-            log.debug(f"LOGIN (password): no such user ({user})")
+            log.debug(f"AUTH (password): no such user ({user})")
             raise AuthException(f"no such user: {user}", 401)
         if not self.check_password(pwd, ref):
-            log.debug(f"LOGIN (password): invalid password for {user}")
+            log.debug(f"AUTH (password): invalid password for {user}")
             raise AuthException(f"invalid password for {user}", 401)
+        return user
+
+    #
+    # FLASK HTTP AUTH
+    #
+    # Use flask_httpauth implementation for basic & digest authentication.
+    #
+    def _get_http_auth(self):
+        assert self._http_auth is not None
+        auth = self._http_auth.get_auth()
+        log.debug(f"auth = {auth}")
+        password = self._http_auth.get_auth_password(auth)
+        log.debug(f"password = {password}")
+        if self._http_auth.authenticate(auth, password):
+            return auth.username
+        log.debug("AUTH (http-*): bad authentication")
+        raise AuthException("failed HTTP authentication", 401)
 
     #
     # HTTP BASIC AUTH
@@ -457,12 +580,20 @@ class FlaskSimpleAuth:
         assert request.remote_user is None
         auth = request.headers.get("Authorization", None)
         log.debug(f"auth: {auth}")
-        if auth is None or auth[:6] != "Basic ":
-            log.debug(f"LOGIN (basic): unexpected auth {auth}")
-            raise AuthException("missing or unexpected authorization header", 401)
-        user, pwd = b64.b64decode(auth[6:]).decode().split(':', 1)
-        self._check_password(user, pwd)
-        return user
+        if auth is None:
+            log.debug("AUTH (basic): missing authorization header")
+            raise AuthException("missing authorization header", 401)
+        if auth[:6] != "Basic ":
+            log.debug(f"AUTH (basic): unexpected auth \"{auth}\"")
+            raise AuthException("unexpected authorization header", 401)
+        try:
+            user, pwd = b64.b64decode(auth[6:]).decode().split(':', 1)
+        except Exception as e:
+            log.debug(f"AUTH (basic): error while decoding auth \"{auth}\" ({e})")
+            raise AuthException("decoding error on authorization header", 401)
+        finally:
+            self._check_password(user, pwd)
+            return user
 
     #
     # HTTP PARAM AUTH
@@ -484,6 +615,13 @@ class FlaskSimpleAuth:
         self._check_password(user, pwd)
         return user
 
+    # basic or param auth
+    def _get_password_auth(self):
+        try:
+            return self._get_basic_auth()
+        except AuthException:  # failed, let's try param
+            return self._get_param_auth()
+
     #
     # TOKEN AUTH
     #
@@ -532,11 +670,13 @@ class FlaskSimpleAuth:
         return jwt.encode({"exp": exp, "sub": user, "aud": realm},
                           secret, algorithm=self._algo)
 
-    def create_token(self, user):
+    def create_token(self, user: str = None):
         """Create a new token for user depending on the configuration."""
-        assert self._type is not None
+        assert self._token is not None
+        if user is None:
+            user = self.get_user()
         realm, delay = self._realm, self._delay
-        if self._type == "fsa":
+        if self._token == "fsa":
             return self._get_fsa_token(realm, user, delay, self._secret)
         else:
             return self._get_jwt_token(realm, user, delay, self._sign)
@@ -547,20 +687,20 @@ class FlaskSimpleAuth:
         realm, user, limit, sig = token.split(':', 3)
         # check realm
         if realm != self._realm:
-            log.debug(f"LOGIN (token): unexpected realm {realm}")
+            log.debug(f"AUTH (fsa token): unexpected realm {realm}")
             raise AuthException(f"unexpected realm: {realm}", 401)
         # check signature
         ref = self._cmp_sig(f"{realm}:{user}:{limit}", self._secret)
         if ref != sig:
-            log.debug("LOGIN (token): invalid signature")
-            raise AuthException("invalid jsa auth token signature", 401)
+            log.debug("AUTH (fsa token): invalid signature")
+            raise AuthException("invalid fsa auth token signature", 401)
         # check limit with a grace time
         now = self._timestamp(dt.datetime.utcnow() - dt.timedelta(minutes=self._grace))
         if now > limit:
-            log.debug("LOGIN (token): token {token} has expired")
-            raise AuthException("expired jsa auth token", 401)
+            log.debug("AUTH (fsa token): token {token} has expired")
+            raise AuthException("expired fsa auth token", 401)
         # all is well
-        return user
+        return user, limit
 
     # jwt authentication can be expensive, especially with pubkey-signatures
     # so use a cache to keep track of already used tokens
@@ -573,10 +713,10 @@ class FlaskSimpleAuth:
             exp = dt.datetime.fromtimestamp(data['exp'])
             return data['sub'], exp
         except jwt.ExpiredSignatureError:
-            log.debug(f"LOGIN (token): token {token} has expired")
+            log.debug(f"AUTH (jwt token): token {token} has expired")
             raise AuthException("expired jwt auth token", 401)
         except Exception as e:
-            log.debug(f"LOGIN (token): invalide token ({e})")
+            log.debug(f"AUTH (jwt token): invalide token ({e})")
             raise AuthException("invalid jwt token", 401)
 
     def _get_jwt_token_auth(self, token):
@@ -584,28 +724,37 @@ class FlaskSimpleAuth:
         # recheck token expiration
         now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
         if now > exp:
-            log.debug(f"LOGIN (token): token {token} has expired")
+            log.debug(f"AUTH (jwt token): token {token} has expired")
             raise AuthException("expired jwt auth token", 401)
-        return user
+        return user, self._timestamp(exp)
 
-    def _get_token_auth(self, token):
-        log.debug(f"checking token: {token}")
+    def _get_token_auth_exp(self, token):
+        log.debug(f"token: {token}")
+        if token is None or token == "":
+            log.debug("AUTH (token): no token")
+            raise AuthException("missing auth token", 401)
         return \
-            self._get_fsa_token_auth(token) if self._type == "fsa" else \
+            self._get_fsa_token_auth(token) if self._token == "fsa" else \
             self._get_jwt_token_auth(token)
 
-    def _get_password_auth(self):
-        try:
-            return self._get_basic_auth()
-        except AuthException:  # failed, let's try param
-            return self._get_param_auth()
+    def _get_token_auth(self, token):
+        return self._get_token_auth_exp(token)[0]
+
+    # a cookie is just another option for storing tokens
+    def _get_cookie_auth(self):
+        return self._get_token_auth(request.cookies[self._name]) \
+            if self._name in request.cookies else None
 
     # map auth types to their functions
     _FSA_AUTH = {
         "basic": _get_basic_auth,
         "param": _get_param_auth,
         "password": _get_password_auth,
-        "fake": _get_fake_auth
+        "fake": _get_fake_auth,
+        "http-basic": _get_http_auth,
+        "http-digest": _get_http_auth,
+        "digest": _get_http_auth,
+        "http-token": _get_http_auth,
     }
 
     def get_user(self):
@@ -627,19 +776,27 @@ class FlaskSimpleAuth:
         elif a == "httpd":
             self._user = request.remote_user
 
-        elif a in ("fake", "basic", "param", "password", "token"):
+        elif a in ("fake", "basic", "param", "password", "token", "http-basic",
+                   "http-digest", "http-token", "digest"):
 
             # check for token
-            if self._type is not None:
-                params = request.values if request.json is None else request.json
-                if self._name is None:
+            # FIXME http-token?
+            if self._token is not None:
+                if self._carrier == "bearer":
                     auth = request.headers.get("Authorization", None)
-                    if auth is not None and auth[:7] == "Bearer ":
-                        self._user = self._get_token_auth(auth[7:])
-                else:
+                    if auth is not None:
+                        slen = len(self._name) + 1
+                        if auth[:slen] == f"{self._name} ":
+                            self._user = self._get_token_auth(auth[slen:])
+                    # else we ignore… maybe it will be resolved later
+                elif self._carrier == "cookie":
+                    self._user = self._get_cookie_auth()
+                elif self._carrier == "param":
+                    params = request.values if request.json is None else request.json
                     token = params.get(self._name, None)
                     if token is not None:
                         self._user = self._get_token_auth(token)
+                # else: cannot get there
 
             # else try other schemes
             if self._user is None:
@@ -659,6 +816,14 @@ class FlaskSimpleAuth:
         """Return current authenticated user, if any."""
         return self._user
 
+    def clear_caches(self):
+        """Clear internal caches."""
+        self._get_jwt_token_auth_real.cache_clear()
+        if self._get_user_pass is not None:
+            self._get_user_pass.cache_clear()
+        if self._user_in_group is not None:
+            self._user_in_group.cache_clear()
+
     #
     # authorize internal decorator
     #
@@ -681,7 +846,7 @@ class FlaskSimpleAuth:
                 self._need_authorization = False
                 # shortcuts
                 if NONE in groups or None in groups:
-                    return "", 403
+                    return Resp("", 403)
                 if ANY in groups:
                     return fun(*args, **kwargs)
                 # get user if needed
@@ -691,20 +856,20 @@ class FlaskSimpleAuth:
                         try:
                             self._user = self.get_user()
                         except AuthException:
-                            return "", 401
+                            return Resp("", 401)
                     else:
-                        return "", 401
+                        return Resp("", 401)
                 if self._user is None:
-                    return "", 401
+                    return Resp("", 401)
                 # shortcut for authenticated users
                 if ALL in groups:
                     return fun(*args, **kwargs)
                 # check against all authorized groups/roles
-                for g in groups:
-                    if self._user_in_group(self._user, g):
+                for r in groups:
+                    if self._user_in_group(self._user, r):
                         return fun(*args, **kwargs)
                 # else no matching group
-                return "", 403
+                return Resp("", 403)
 
             return wrapper
 
@@ -749,7 +914,7 @@ class FlaskSimpleAuth:
 
                 # this cannot happen under normal circumstances
                 if self._need_authorization and self._check:
-                    return "missing authorization check", 500
+                    return Resp("missing authorization check", 500)
 
                 # translate request parameters to named function parameters
                 params = request.values if request.json is None else request.json
@@ -760,15 +925,15 @@ class FlaskSimpleAuth:
                             try:
                                 kwargs[p] = typing(params[p])
                             except Exception as e:
-                                return f"type error on HTTP parameter {p} ({e})", 400
+                                return Resp(f"type error on HTTP parameter \"{p}\" ({e})", 400)
                         else:
                             if required is None:
                                 if p in defaults:
                                     kwargs[p] = defaults[p]
                                 else:
-                                    return f"missing HTTP parameter {p}", 400
+                                    return Resp(f"missing HTTP parameter \"{p}\"", 400)
                             elif required:
-                                return f"missing HTTP parameter {p}", 400
+                                return f"missing HTTP parameter \"{p}\"", 400
                             else:
                                 kwargs[p] = defaults.get(p, None)
                     else:
@@ -777,7 +942,7 @@ class FlaskSimpleAuth:
                             try:
                                 kwargs[p] = typing(kwargs[p])
                             except Exception as e:
-                                return f"type error on path parameter {p}: ({e})", 404
+                                return Resp(f"type error on path parameter \"{p}\": ({e})", 404)
 
                 # possibly add others, without shadowing already provided ones
                 if allparams:
@@ -822,7 +987,6 @@ class FlaskSimpleAuth:
                 if ":" not in spec and spec in sig.parameters:
                     t = typeof(sig.parameters[spec])
                     # Flask supports 5 types, with string the default?
-                    # FIXME how to handle path?
                     if t in (int, float, UUID, path):
                         splits[i] = f"{t.__name__.lower()}:{spec}>{remainder}"
                     else:
@@ -836,6 +1000,9 @@ class FlaskSimpleAuth:
 
     def route(self, rule, **options):
         """Extended `route` decorator provided by the extension."""
+        if "authorize" not in options:
+            log.warning(f"missing authorize on route \"{rule}\" makes it 403 Forbidden")
+
         def decorate(fun):
             self.add_url_rule(rule, view_func=fun, **options)
         return decorate
